/**************************************
****   INPUT.C  -  Input devices   ****
**************************************/

//-- Include Files -----------------------------------------------------------
#include <stdio.h>
#include <allegro.h>
#include "../config/config.h"
#include "../gui/gui.h"
#include "macro.h"

//-- Structures --------------------------------------------------------------
typedef struct {
	unsigned char	key_name[16];
	unsigned int	key_scancode;
} KEYDEF;

typedef struct {
	unsigned char	joy_name[28];
	unsigned int	joy_id;
} JOYDEF;

extern void not_implemented(void);

//-- Private Variables -------------------------------------------------------
JOYDEF	joydefs[11] = {
	{"No Joystick",              JOY_TYPE_NONE},
	{"4 Buttons",                JOY_TYPE_4BUTTON},
	{"6 Buttons",                JOY_TYPE_6BUTTON},
	{"8 Buttons",                JOY_TYPE_8BUTTON},
	{"CH FlightStick Pro",       JOY_TYPE_FSPRO},
	{"Wingman Extreme",          JOY_TYPE_WINGEX},
	{"MS Sidewinder",            JOY_TYPE_SIDEWINDER},
	{"Gravis Gamepad Pro",       JOY_TYPE_GAMEPAD_PRO},
	{"SNES Pad LPT1",            JOY_TYPE_SNESPAD_LPT1},
	{"SNES Pad LPT2",            JOY_TYPE_SNESPAD_LPT2},
	{"SNES Pad LPT3",            JOY_TYPE_SNESPAD_LPT3}
	};
	
KEYDEF	keydefs[96] = {
	{"KEY_NONE",               0},
	{"KEY_ESC",                1},
	{"KEY_1",                  2},
	{"KEY_2",                  3},
	{"KEY_3",                  4},
	{"KEY_4",                  5},
	{"KEY_5",                  6},
	{"KEY_6",                  7},
	{"KEY_7",                  8},
	{"KEY_8",                  9},
	{"KEY_9",                 10},
	{"KEY_0",                 11},
	{"KEY_MINUS",             12},
	{"KEY_EQUALS",            13},
	{"KEY_BACKSPACE",         14},
	{"KEY_TAB",               15},
	{"KEY_Q",                 16},
	{"KEY_W",                 17},
	{"KEY_E",                 18},
	{"KEY_R",                 19},
	{"KEY_T",                 20},
	{"KEY_Y",                 21},
	{"KEY_U",                 22},
	{"KEY_I",                 23},
	{"KEY_O",                 24},
	{"KEY_P",                 25},
	{"KEY_OPENBRACE",         26},
	{"KEY_CLOSEBRACE",        27},
	{"KEY_ENTER",             28},
	{"KEY_CONTROL",           29},
	{"KEY_LCONTROL",          29},
	{"KEY_A",                 30},
	{"KEY_S",                 31},
	{"KEY_D",                 32},
	{"KEY_F",                 33},
	{"KEY_G",                 34},
	{"KEY_H",                 35},
	{"KEY_J",                 36},
	{"KEY_K",                 37},
	{"KEY_L",                 38},
	{"KEY_COLON",             39},
	{"KEY_QUOTE",             40},
	{"KEY_TILDE",             41},
	{"KEY_LSHIFT",            42},
	{"KEY_BACKSLASH",         43},
	{"KEY_Z",                 44},
	{"KEY_X",                 45},
	{"KEY_C",                 46},
	{"KEY_V",                 47},
	{"KEY_B",                 48},
	{"KEY_N",                 49},
	{"KEY_M",                 50},
	{"KEY_COMMA",             51},
	{"KEY_STOP",              52},
	{"KEY_SLASH",             53},
	{"KEY_RSHIFT",            54},
	{"KEY_ASTERISK",          55},
	{"KEY_ALT",               56},
	{"KEY_SPACE",             57},
	{"KEY_CAPSLOCK",          58},
	{"KEY_F1",                59},
	{"KEY_F2",                60},
	{"KEY_F3",                61},
	{"KEY_F4",                62},
	{"KEY_F5",                63},
	{"KEY_F6",                64},
	{"KEY_F7",                65},
	{"KEY_F8",                66},
	{"KEY_F9",                67},
	{"KEY_F10",               68},
	{"KEY_NUMLOCK",           69},
	{"KEY_SCRLOCK",           70},
	{"KEY_HOME",              71},
	{"KEY_UP",                72},
	{"KEY_PGUP",              73},
	{"KEY_MINUS_PAD",         74},
	{"KEY_LEFT",              75},
	{"KEY_5_PAD",             76},
	{"KEY_RIGHT",             77},
	{"KEY_PLUS_PAD",          78},
	{"KEY_END",               79},
	{"KEY_DOWN",              80},
	{"KEY_PGDN",              81},
	{"KEY_INSERT",            82},
	{"KEY_DEL",               83},
	{"KEY_PRTSCR",            84},
	{"KEY_F11",               87},
	{"KEY_F12",               88},
	{"KEY_LWIN",              91},
	{"KEY_RWIN",              92},
	{"KEY_MENU",              93},
	{"KEY_PAD",              100},
	{"KEY_RCONTROL",         120},
	{"KEY_ALTGR",            121},
	{"KEY_SLASH2",           122},
	{"KEY_PAUSE",            123}
};

char	*key_getter(int index, int *list_size)
{
	if (index < 0) {
		*list_size = 96;
		return NULL;
	}
	
	return (keydefs[index].key_name + 4);
}

char	*joy_getter(int index, int *list_size)
{
	if (index < 0) {
		*list_size = 11;
		return NULL;
	}
	
	return (joydefs[index].joy_name);
}

char	*act1_getter(int index, int *list_size)
{
	static char *act1_str[] = {
		"Up",
		"Down",
		"Left",
		"Right",
		"A",
		"B",
		"C",
		"D",
		"Start",
		"Select",
		"Macro01",		
		"Macro02",		
		"Macro03",		
		"Macro04",		
		"Macro05",		
		"Macro06",		
		"Macro07",		
		"Macro08",		
		"Macro09",		
		"Macro10",		
		"Macro11",		
		"Macro12",		
		"Macro13",		
		"Macro14",		
		"Macro15",		
		"Macro16"
	};
	
   if (index < 0) {
      *list_size = 26;
      return NULL;
   }

   return act1_str[index]; 
}	

char	*act2_getter(int index, int *list_size)
{
	static char *act2_str[] = {
		"A",
		"B",
		"C",
		"D",
		"Start",
		"Select",
		"Macro01",		
		"Macro02",		
		"Macro03",		
		"Macro04",		
		"Macro05",		
		"Macro06",		
		"Macro07",		
		"Macro08",		
		"Macro09",		
		"Macro10",		
		"Macro11",		
		"Macro12",		
		"Macro13",		
		"Macro14",		
		"Macro15",		
		"Macro16"
	};
	
   if (index < 0) {
      *list_size = 22;
      return NULL;
   }

   return act2_str[index]; 
}

char	*btn_getter(int index, int *list_size)
{
	static char *btn_str = "But. 00";
	
	if (index < 0) {
		*list_size = 22;
		return NULL;
	}
	
	sprintf(btn_str, "But. %02d", index);
	return btn_str;
}

DIALOG kbd_setup_dlg[] =
{
   /* (dialog proc)   (x)  (y)  (w)  (h)    (fg)    (bg)  (key)  (flags)   (d1)  (d2)            (dp) */
 {d_neo_border_proc,   0,   0, 224, 225, 0x0000, 0x0000,      0,      0,     0,    0,                NULL},
 {d_neo_fbox_proc,     2,   2, 220,   8, 0x0000, 0x10E3,      0,      0,     0,    0,                NULL},
 {d_text_proc,        10,   3,   0,   0, 0xFFFF, 0x10E3,      0,      0,     0,    0,    "Keyboard Setup"},
 {d_neo_gbox_proc,     5,  20, 214,  70, 0xC71C, 0x328A,      0,      0,     0,    0,        "Player One"},
 {d_neo_list_proc,    10,  26,  83,  59, 0xC71C, 0x328A,      0,      0,     0,    0,          key_getter},
 {d_neo_button_proc,  98,  38,  40,  15, 0x0000, 0x0000,      0, D_EXIT,     0,    0,             "<<Get"},
 {d_neo_button_proc,  98,  58,  40,  15, 0x0000, 0x0000,      0, D_EXIT,     0,    0,             "Set>>"},
 {d_neo_list_proc,   143,  26,  71,  59, 0xC71C, 0x328A,      0,      0,     0,    0,         act1_getter},

 {d_neo_button_proc,  79, 100,  65,  15, 0x0000, 0x0000,    'e', D_EXIT,     0,    0,         "&Exchange"},

 {d_neo_gbox_proc,     5, 125, 214,  70, 0xC71C, 0x328A,      0,      0,     0,    0,        "Player Two"},
 {d_neo_list_proc,    10, 131,  83,  59, 0xC71C, 0x328A,      0,      0,     0,    0,          key_getter},
 {d_neo_button_proc,  98, 143,  40,  15, 0x0000, 0x0000,      0, D_EXIT,     0,    0,             "<<Get"},
 {d_neo_button_proc,  98, 163,  40,  15, 0x0000, 0x0000,      0, D_EXIT,     0,    0,             "Set>>"},
 {d_neo_list_proc,   143, 131,  71,  59, 0xC71C, 0x328A,      0,      0,     0,    0,         act1_getter},
 
 {d_neo_button_proc,  42, 203, 65,  15, 0x0000, 0x0000,     'o', D_EXIT,     0,    0,               "&OK"},
 {d_neo_button_proc, 117, 203, 65,  15, 0x0000, 0x0000,     'm', D_EXIT,     0,    0,        "&Macrokeys"},

 {NULL,                0,   0,   0,   0, 0x0000, 0x0000,      0,      0,     0,    0,                NULL}
};

DIALOG joy_setup_dlg[] =
{
   /* (dialog proc)   (x)  (y)  (w)  (h)    (fg)    (bg)  (key)  (flags)   (d1)  (d2)            (dp) */
 {d_neo_border_proc,   0,   0, 319, 200, 0x0000, 0x0000,      0,      0,     0,    0,                NULL},
 {d_neo_fbox_proc,     2,   2, 315,   8, 0x0000, 0x10E3,      0,      0,     0,    0,                NULL},
 {d_text_proc,        10,   3,   0,   0, 0xFFFF, 0x10E3,      0,      0,     0,    0,    "Joystick Setup"},
 {d_neo_gbox_proc,     5,  20, 190,  70, 0xC71C, 0x328A,      0,      0,     0,    0,        "Player One"},
 {d_neo_list_proc,    10,  26,  65,  59, 0xC71C, 0x9596,      0,      0,     0,    0,          btn_getter},
 {d_neo_button_proc,  80,  38,  40,  15, 0x0000, 0x0000,      0, D_EXIT,     0,    0,             "<<Get"},
 {d_neo_button_proc,  80,  58,  40,  15, 0x0000, 0x0000,      0, D_EXIT,     0,    0,             "Set>>"},
 {d_neo_list_proc,   125,  26,  65,  59, 0xC71C, 0x9596,      0,      0,     0,    0,         act2_getter},

 {d_neo_gbox_proc,     5, 100, 190,  70, 0xC71C, 0x328A,      0,      0,     0,    0,        "Player Two"},
 {d_neo_list_proc,    10, 106,  65,  59, 0xC71C, 0x9596,      0,      0,     0,    0,          btn_getter},
 {d_neo_button_proc,  80, 118,  40,  15, 0x0000, 0x0000,      0, D_EXIT,     0,    0,             "<<Get"},
 {d_neo_button_proc,  80, 138,  40,  15, 0x0000, 0x0000,      0, D_EXIT,     0,    0,             "Set>>"},
 {d_neo_list_proc,   125, 106,  65,  59, 0xC71C, 0x9596,      0,      0,     0,    0,         act2_getter},

 {d_neo_gbox_proc,   200,  20, 115, 150, 0xC71C, 0x328A,      0,      0,     0,    0,     "Joystick Type"},
 {d_neo_list_proc,   205,  26, 105, 119, 0xC71C, 0x9596,      0,      0,     0,    0,          joy_getter},
 {d_neo_button_proc, 205, 150, 105,  15, 0x0000, 0x0000,    'c', D_EXIT,     0,    0,           "&Change"},
 
 {d_neo_button_proc,  55, 180, 100,  15, 0x0000, 0x0000,    'o', D_EXIT,     0,    0,               "&OK"},
 {d_neo_button_proc, 165, 180, 100,  15, 0x0000, 0x0000,    'm', D_EXIT,     0,    0, "Select &Macrokeys"},

 {NULL,                0,   0,   0,   0, 0x0000, 0x0000,      0,      0,     0,    0,                NULL}
};

static int	macro_number = 0;
static int	macro_pos = 0;
static int	macro_counter;

//-- Global Variables --------------------------------------------------------
int	macro_playing = 0;

void	keyboard_setup(void);
				
//----------------------------------------------------------------------------
unsigned char	read_player1(void)
{
	unsigned char	k = 0xFF;
	int		i = 0;

	if ( !macro_playing )
	{
		if (key[kbd_p1.up]||joy[0].stick[0].axis[1].d1)
			k &= 0xFE;

		if (key[kbd_p1.down]||joy[0].stick[0].axis[1].d2)
			k &= 0xFD;

		if (key[kbd_p1.left]||joy[0].stick[0].axis[0].d1)
			k &= 0xFB;

		if (key[kbd_p1.right]||joy[0].stick[0].axis[0].d2)
			k &= 0xF7;

		if (key[kbd_p1.a]||joy[0].button[joy_p1.a].b)
			k &= 0xEF;

		if (key[kbd_p1.b]||joy[0].button[joy_p1.b].b)
			k &= 0xDF;

		if (key[kbd_p1.c]||joy[0].button[joy_p1.c].b)
			k &= 0xBF;

		if (key[kbd_p1.d]||joy[0].button[joy_p1.d].b)
			k &= 0x7F;
	
		if (k==0xFF)
		{
			while(macro_list[i].scancode != 0)
			{
				if (key[macro_list[i].scancode] != 0)
				{
					macro_playing = 1;
					macro_number = i;
					macro_pos = 0;
					macro_counter = 5;
				}
			
				i++;
			}
		}
	}

	if ( macro_playing )
	{
		if (macro_list[macro_number].macro[macro_pos] == 0xFF)
			macro_playing = 0;
		else
		{
			k = ~macro_list[macro_number].macro[macro_pos];
			if (macro_counter != 0)
				macro_counter--;
			else
			{
				macro_counter = 5;
				macro_pos++;
			}
		}
	}

	return	k;
}

//----------------------------------------------------------------------------
unsigned char	read_player2(void)
{
	unsigned char	k = 0xFF;

	if (key[kbd_p2.up]||joy[1].stick[0].axis[1].d1)
		k &= 0xFE;

	if (key[kbd_p2.down]||joy[1].stick[0].axis[1].d2)
		k &= 0xFD;

	if (key[kbd_p2.left]||joy[1].stick[0].axis[0].d1)
		k &= 0xFB;

	if (key[kbd_p2.right]||joy[1].stick[0].axis[0].d2)
		k &= 0xF7;

	if (key[kbd_p2.a]||joy[1].button[joy_p2.a].b)
		k &= 0xEF;

	if (key[kbd_p2.b]||joy[1].button[joy_p2.b].b)
		k &= 0xDF;

	if (key[kbd_p2.c]||joy[1].button[joy_p2.c].b)
		k &= 0xBF;

	if (key[kbd_p2.d]||joy[1].button[joy_p2.d].b)
		k &= 0x7F;
	
	return	k;
}

//----------------------------------------------------------------------------
char	read_pl12_startsel(void)
{
	char	k = 0x0F;
	
	if (key[kbd_p1.start]||joy[0].button[joy_p1.start].b)
		k &= 0xFE;
		
	if (key[kbd_p1.select]||joy[0].button[joy_p1.select].b)
		k &= 0xFD;

	if (key[kbd_p2.start]||joy[1].button[joy_p2.start].b)
		k &= 0xFB;
		
	if (key[kbd_p2.select]||joy[1].button[joy_p2.select].b)
		k &= 0xF7;

	return k;
}

//----------------------------------------------------------------------------
unsigned int	input_identify(unsigned char *key_name)
{
	unsigned int	i;
	
	for(i=0;i<96;i++)
		if (strcmp((char *)key_name, (char *)keydefs[i].key_name) == 0)
			return keydefs[i].key_scancode;
	
	return 0;
}

//----------------------------------------------------------------------------
unsigned char	*input_whois(unsigned int key_scancode)
{
	unsigned int	i;
	
	for(i=0;i<96;i++)
		if (keydefs[i].key_scancode == key_scancode)
			return keydefs[i].key_name;
	
	return	NULL;
}

void	keyboard_setup(void)
{
	int	ret;
	
	
	centre_dialog(kbd_setup_dlg);
	
	do
	{
		vsync();
		ret = do_dialog(kbd_setup_dlg, 0);
		
		if (ret==5)
			kbd_setup_dlg[4].d1 = val_to_index(config_get_var_kbd(&kbd_p1, kbd_setup_dlg[7].d1));
			
		if (ret==6)
			config_set_var_kbd(&kbd_p1, kbd_setup_dlg[7].d1, keydefs[kbd_setup_dlg[4].d1].key_scancode);

		if (ret==8)
		{
			CONFIG	temp;
			
			memcpy((void *)&temp, (void *)&kbd_p1, sizeof(CONFIG));
			memcpy((void *)&kbd_p1, (void *)&kbd_p2, sizeof(CONFIG));
			memcpy((void *)&kbd_p2, (void *)&temp, sizeof(CONFIG));
			
			neo_alert1("Notification", NULL, "Exchanged P1 <-> P2 configurations.", NULL, "&Ok", 'o');
		}
	
		if (ret==11)
			kbd_setup_dlg[10].d1 = val_to_index(config_get_var_kbd(&kbd_p2, kbd_setup_dlg[13].d1));

		if (ret==12)
			config_set_var_kbd(&kbd_p2, kbd_setup_dlg[13].d1, keydefs[kbd_setup_dlg[10].d1].key_scancode);
			
		if (ret==15)
			not_implemented();
		
		
	} while(ret != 14);

	config_save_kbd_conf(&kbd_p1, "keyboard1");
	config_save_kbd_conf(&kbd_p2, "keyboard2");

}

void joystick_setup(void)
{
	int	ret, i;
	
	centre_dialog(joy_setup_dlg);

	ret = get_config_id("joystick", "joytype", JOY_TYPE_NONE);

	for(i=0;i<11;i++)
	{
		if (ret == joydefs[i].joy_id)
			break;
	}

	joy_setup_dlg[14].d1 = i;
	
	do {
		vsync();	
		ret = do_dialog(joy_setup_dlg, 0);
		
		if (ret==5)
			joy_setup_dlg[4].d1 = config_get_var_joy(&joy_p1, joy_setup_dlg[7].d1);
		
		if (ret==6)
			config_set_var_joy(&joy_p1, joy_setup_dlg[7].d1, joy_setup_dlg[4].d1);

		if (ret==10)
			joy_setup_dlg[9].d1 = config_get_var_joy(&joy_p2, joy_setup_dlg[12].d1);
		
		if (ret==11)
			config_set_var_joy(&joy_p2, joy_setup_dlg[12].d1, joy_setup_dlg[9].d1);
		
		if (ret==15)
		{
			int	val;
			char	msg1[80];
			char	msg2[80];
			
			remove_joystick();
			val = install_joystick(joydefs[joy_setup_dlg[14].d1].joy_id);
			
			if (val != 0)
				neo_alert1("ERROR", NULL, "Joystick initialization failed! (is it plugged?)", NULL, "&Abort", 'a');
			else
			{
				for (val=0; val<num_joysticks; val++) {
					if (joy[val].flags & JOYFLAG_CALIBRATE)
						sprintf(msg1, "Please calibrate joystick #%d", val);
					
					while (joy[val].flags & JOYFLAG_CALIBRATE) {
						sprintf(msg2, "%s, and press OK.", calibrate_joystick_name(val));

						rest(10);
						neo_alert1(msg1, NULL, msg2, NULL, "&Ok", 'o');
	 
		 				calibrate_joystick(val);
					
						rest(10);
					}
				}
			}
			
			neo_alert1("OK", NULL, "Joystick setup complete !", NULL, "&Ok", 'o');
			save_joystick_data(NULL);
		}
		
		if (ret==17)
			not_implemented();
		
	} while(ret != 16);

	config_save_joy_conf(&joy_p1, "joystick1");
	config_save_joy_conf(&joy_p2, "joystick2");

}
