/**************************************
****  NEOCD.C  - Main Source File  ****
**************************************/

//-- Include Files -----------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <allegro.h>
#include <dos.h>
#include "cdaudio/cdaudio.h"
#include "cdrom/cdrom.h"
#include "mc68000/68kregs.h"
#include "memory/memory.h"
#include "video/video.h"
#include "config/config.h"
#include "input/input.h"
#include "input/macro.h"
#include "gui/gui.h"
#include "gui/rsc.h"
#include "winlow/winlow.h"

// Cut off Allegro size...
DECLARE_DIGI_DRIVER_LIST()
DECLARE_MIDI_DRIVER_LIST()
BEGIN_GFX_DRIVER_LIST
	GFX_DRIVER_VESA3
	GFX_DRIVER_VESA2L
	GFX_DRIVER_VESA2B
	GFX_DRIVER_VESA1
END_GFX_DRIVER_LIST

// 12MHz/224lines/60fps = 893 cycles per line
#define	CYCLES_PER_LINE	893

// Version, date & time to display on startup
#define VERSION1 "NEOGEO-CD Emulator v0.45"
#define VERSION2 "("__DATE__" "__TIME__")"

//-- Global Variables --------------------------------------------------------
char			*neogeo_rom_memory = NULL;
char			*neogeo_prg_memory = NULL;
char			*neogeo_fix_memory = NULL;
char			*neogeo_spr_memory = NULL;
char			global_error[80];
int				neogeo_vline;
int				neogeo_68k_cycles;
int				neogeo_last_command;
unsigned char	neogeo_memorycard[8192];
int				neogeo_prio_mode = 0;
DATAFILE		*rsc;
int				neogeo_ipl_done = 0;
int				windows_present;
int				windows_hostile_mode = 0;

//-- 68K Core related stuff --------------------------------------------------
int					mame_debug = 0;
char				*OP_ROM;
char				*OP_RAM;
int					previouspc = 0;
int					ophw = 0;
int					cur_mrhard = 0;
extern int			MC68000_ICount;
extern Registers68K	regs;
extern void			M68KRESET(void);	// 68K core init, not reset
extern int			M68KRUN(int);
extern void			initialize_memmap(void);

//-- Function Prototypes -----------------------------------------------------
void	neogeo_init(void);
void	neogeo_reset(void);
void	neogeo_hreset(void);
void	neogeo_shutdown(void);
void	MC68000_Cause_Interrupt(int);
void	neogeo_exception(void);
void	neogeo_run(void);
void	draw_main(void);
void	neogeo_quit(void);
void	not_implemented(void);
void	neogeo_machine_settings(void);

//-- Main screen buttons -----------------------------------------------------
typedef void (*CALLBACK)(void);

typedef struct {
	int			x1;
	int			y1;
	int			x2;
	int			y2;
	int			nrm;
	int			sel;
	CALLBACK	func;
} MAIN_ITEM;

MAIN_ITEM	main_buttons[] = {
	{ 212, 101, 283, 114, BT_RUN1,    BT_RUN2,    neogeo_run },
	{ 212, 116, 283, 129, BT_VSETUP1, BT_VSETUP2, video_setup },
	{ 212, 131, 283, 144, BT_ASETUP1, BT_ASETUP2, not_implemented },
	{ 212, 146, 283, 159, BT_JSETUP1, BT_JSETUP2, joystick_setup },
	{ 212, 161, 283, 174, BT_KSETUP1, BT_KSETUP2, keyboard_setup },
	{ 212, 176, 283, 189, BT_SYSSET1, BT_SYSSET2, neogeo_machine_settings },
	{ 212, 191, 283, 204, BT_QUIT1,   BT_QUIT2,   neogeo_quit }
};

#define NB_ELEMS	7

char *nation_getter(int index, int *list_size)
{
	static char *nats[] =
	{
		"Japanese", "American", "European"
	};
	
	if (index < 0)
	{
		*list_size = 3;
		return NULL;
	}
	else
		return nats[index];
}

char *drive_getter(int index, int *list_size)
{
	static char *aut = "Detect";
	static char *drv = "X:";
	
	if (index < 0)
	{
		*list_size = nb_of_drives + 1;
		return NULL;
	}
	
	if (index == 0)
		return aut;
	
	drv[0] = drive_list[index-1] + 65;
	
	return drv;
}


DIALOG machine_settings[] =
{
   /* (dialog proc)   (x)  (y)  (w)  (h)    (fg)    (bg)  (key)  (flags)   (d1)  (d2)            (dp) */
 {d_neo_border_proc,   0,   0, 170, 190, 0x0000, 0x0000,      0,      0,     0,    0,                NULL},
 {d_neo_fbox_proc,     2,   2, 166,   8, 0x0000, 0x10E3,      0,      0,     0,    0,                NULL},
 {d_text_proc,        10,   3,   0,   0, 0xFFFF, 0x10E3,      0,      0,     0,    0,  "Machine Settings"},

 {d_neo_gbox_proc,    10,  20, 150,  40, 0xC71C, 0x328A,      0,      0,     0,    0,       "Nationality"},
 {d_neo_list_proc,    15,  25, 140,  30, 0xC71C, 0x9596,      0,      0,     0,    0,       nation_getter},

 {d_neo_button_proc,  10,  65,  60,  15, 0x0000, 0x0000,    'r', D_EXIT,     0,    0,            "&Reset"},
 {d_neo_button_proc,  75,  65,  85,  15, 0x0000, 0x0000,    'f', D_EXIT,     0,    0,     "&Force Reload"},

 {d_neo_gbox_proc,    10,  90,  70,  50, 0xC71C, 0x328A,      0,      0,     0,    0,            "CD-ROM"},
 {d_neo_gbox_proc,    90,  90,  70,  50, 0xC71C, 0x328A,      0,      0,     0,    0,          "CD-Audio"},

 {d_neo_list_proc,    15,  95,  60,  40, 0xC71C, 0x9596,      0,      0,     0,    0,        drive_getter},
 {d_neo_list_proc,    95,  95,  60,  40, 0xC71C, 0x9596,      0,     32,     0,    0,        drive_getter},

 {d_neo_button_proc,  10, 145,  70,  15, 0x0000, 0x0000,      0, D_EXIT,     0,    0,            "Change"},
 {d_neo_button_proc,  90, 145,  70,  15, 0x0000, 0x0000,      0,     32,     0,    0,            "Change"},

 {d_neo_button_proc,  10, 170,  70,  15, 0x0000, 0x0000,    'a', D_EXIT,     0,    0,            "&About"},
 {d_neo_button_proc,  90, 170,  70,  15, 0x0000, 0x0000,    'o', D_EXIT,     0,    0,               "&OK"},

 {NULL,                0,   0,   0,   0, 0x0000, 0x0000,      0,      0,     1,    1,                NULL}
};

DIALOG about_box[] =
{
 {d_box_proc,          0,   0, 150, 130, 0x0000, 0xFFFF,      0,      0,     0,    0,                NULL},
 {d_icon_proc,         5,   1, 141, 106, 0xFFFF, 0xFFFF,      0, D_EXIT,     0,    0,                NULL},
 {d_text_proc,         4, 107,   0,   0, 0x0000, 0xFFFF,      0,      0,     0,    0,            VERSION1},
 {d_text_proc,        10, 115,   0,   0, 0x0000, 0xFFFF,      0,      0,     0,    0,            VERSION2},
 {NULL,                0,   0,   0,   0, 0x0000, 0x0000,      0,      0,     0,    0,                NULL}
};

//----------------------------------------------------------------------------
int	main(void)
{
	MACROFILE	fm;
	FILE		*fp;
	int			result;
	int			a, b;
	
	atexit(neogeo_shutdown);
	puts(VERSION1);
	puts(VERSION2);
	
	windows_present = winlow_is_windows_present();
	
	if (windows_present)
		puts ("\nRunning under Windows...\n");
	else
		puts("\nRunning under DOS...\n");

	puts("NEOGEO: Allocating memory...");

	printf("PRG (2Mo) ... ");
	neogeo_prg_memory = calloc(1, 2097152);
	if (neogeo_prg_memory==NULL) {
		puts("failed !");
		return 0;
	}
	puts("OK!");

	printf("SPR (4Mo) ... ");
	neogeo_spr_memory = calloc(1, 4194304);
	if (neogeo_spr_memory==NULL) {
		puts("failed !");
		return 0;
	}
	puts("OK!");
	
	printf("ROM (128k) ... ");
	neogeo_rom_memory = calloc(1, 131072);
	if (neogeo_rom_memory==NULL) {
		puts("failed !");
		return 0;
	}
	puts("OK!");

	printf("FIX (128k) ... ");
	neogeo_fix_memory = calloc(1, 131072);
	if (neogeo_fix_memory==NULL) {
		puts("failed !");
		return 0;
	}
	puts("OK!");
	
	// Initialize Memory Mapping
	initialize_memmap();
	
	/* initialise everything */
	set_config_file("neocd.cfg");
	allegro_init();

	windows_hostile_mode = get_config_int(NULL, "windows_hostile", 0);
	
	if ((windows_present)&&(windows_hostile_mode))
		winlow_disable_windows();

	install_keyboard(); 
	install_mouse();
	install_timer();
	set_color_depth(16);

	// Load and Initialize rsc Datafile
	rsc = load_datafile("rsc.dat");
	if (rsc==NULL)
	{
		puts("FATAL: Couldn't find ressource datafile RSC.DAT");
		exit(0);
	}
	fixup_datafile(rsc);

	// Video init
	result = video_init();
	if (!result)
	{
		video_mode = 0;
		puts(global_error);
		return 0;
	}

	font = rsc[FNT].dat;
	set_mouse_sprite(rsc[CURSOR].dat);
	gui_fg_color = 0xFFFF;
	gui_mg_color = 0xC71C;
	gui_bg_color = 0x9596;

	show_mouse(screen);

	fp = fopen("neocd.bin", "rb");
	if (fp==NULL) {
		neo_alert1("Fatal Error", NULL, "Could not load NEOCD.BIN", NULL, "&Abort", 'a');
		return 0;
	}
	fread(neogeo_rom_memory, 1, 131072, fp);
	fclose(fp);
	swab(neogeo_rom_memory, neogeo_rom_memory, 131072);

	fp = fopen("startup.bin", "rb");
	if (fp==NULL) {
		neo_alert1("Fatal Error", NULL, "Could not load STARTUP.BIN.", NULL, "&Abort", 'a');
		return 0;
	}

	fread(neogeo_prg_memory + 0x10F300, 1, 3328, fp);
	fclose(fp);
	swab(neogeo_prg_memory + 0x10F300, neogeo_prg_memory + 0x10F300, 3328);

	// Check BIOS validity
	if (*((short*)(neogeo_rom_memory+0xA822)) != 0x4BF9)
	{
		neo_alert1("Fatal Error", NULL, "Invalid BIOS file.", NULL, "&Abort", 'a');
		return 0;
	}

	/*** Patch BIOS exit ***/
	*((short*)(neogeo_rom_memory+0x55e)) = 0xFABE;
	*((short*)(neogeo_rom_memory+0x560)) = 0x4E75;
	/*** Patch BIOS load files w/ now loading message ***/
	*((short*)(neogeo_rom_memory+0x552)) = 0xFABF;
	*((short*)(neogeo_rom_memory+0x554)) = 0x4E75;
	/*** Patch BIOS load files w/out now loading ***/
	*((short*)(neogeo_rom_memory+0x564)) = 0xFAC0;
	*((short*)(neogeo_rom_memory+0x566)) = 0x4E75;
	/*** Patch BIOS CDROM Check ***/
	*((short*)(neogeo_rom_memory+0xB040)) = 0x4E71;
	*((short*)(neogeo_rom_memory+0xB042)) = 0x4E71;
	/*** Patch BIOS upload command ***/
	*((short*)(neogeo_rom_memory+0x546)) = 0xFAC1;
	*((short*)(neogeo_rom_memory+0x548)) = 0x4E75;

	/*** Full reset, please ***/
	*((short*)(neogeo_rom_memory+0xA87A)) = 0x4239;
	*((short*)(neogeo_rom_memory+0xA87C)) = 0x0010;
	*((short*)(neogeo_rom_memory+0xA87E)) = 0xFDAE;

	/*** Trap exceptions ***/
	*((short*)(neogeo_rom_memory+0xA5B6)) = 0x4AFC;

	fp = fopen("memcard.bin", "rb");
	if (fp!=NULL) {
		fread(neogeo_memorycard, 1, 8192, fp);
		fclose(fp);
	}

	// Initialize Macros
	macro_init();
	
	// Install joystick(s)
	load_joystick_data(NULL);

	// Read config files
	config_read();

	// Initialize CD-DA
	cdda_current_drive = cdrom_current_drive;
	result = cdda_init();
	puts(global_error);
	if (!result)
		return 0;
	cdrom_current_drive = cdda_current_drive;

	// CD-ROM init
	result = cdrom_init();
	if (!result)
	{
		neo_alert1("Fatal Error", NULL, global_error, NULL, "&Abort", 'a');
		return 0;
	}

	neogeo_init();

	draw_main();
	
	while(1)
	{	
		while( !mouse_b&&!keypressed() );

		b = -1;

		if (keypressed())
		{
			a = readkey()&0xDF;
			
			switch(a)
			{
				case 'R':
					b = 0;
					break;
				case 'V':
					b = 1;
					break;
				case 'A':
					b = 2;
					break;
				case 'J':
					b = 3;
					break;
				case 'K':
					b = 4;
					break;
				case 'S':
					b = 5;
					break;
				case 'Q':
					b = 6;
					break;
			}

			if (b>=0)
			{
				scare_mouse();
				blit(rsc[main_buttons[b].sel].dat, screen, 0, 0, main_buttons[b].x1, main_buttons[b].y1, 71, 13);
				unscare_mouse();
			}		
		}
		else
		{
			for(a=0;a<NB_ELEMS;a++)
			{
				if ((mouse_x>=main_buttons[a].x1)&&(mouse_x<=main_buttons[a].x2)&&
					(mouse_y>=main_buttons[a].y1)&&(mouse_y<=main_buttons[a].y2))
				{
					b = a;
					scare_mouse();
					blit(rsc[main_buttons[a].sel].dat, screen, 0, 0, main_buttons[a].x1, main_buttons[a].y1, 71, 13);
					unscare_mouse();
					while( mouse_b );
					break;
				}
			}
		}
	
		if (b>=0)
		{
			main_buttons[b].func();
			vsync();
			draw_main();
		}
	}
}

void	neogeo_init(void)
{
	M68KRESET();
}

void	neogeo_hreset(void)
{
	FILE	*fp;

	// Read game name (used in macro system)
	config_read_name();
	strtrim(config_game_name, config_game_name);

	// Special patch
	if (strcmp(config_game_name, "TEST PROGRAM USA") == 0)
	{
		fp = fopen("patch.prg", "rb");
		fread(neogeo_prg_memory + 0x132000, 1, 112, fp);
		fclose(fp);
		swab(neogeo_prg_memory + 0x132000, neogeo_prg_memory + 0x132000, 112);
	}

	// First time init
	memset((void *)&regs, 0, sizeof(regs));
	regs.pc = 0xc0a822;
	regs.sr = 0x2700;
	regs.sr_h = 0x27;
	regs.a7 = regs.isp = regs.usp = 0x10F300;
	OP_ROM = neogeo_rom_memory - 0xC00000;
	neogeo_vline = 0;
	neogeo_68k_cycles = CYCLES_PER_LINE;
	cpu_writemem24_dword(0x10F6EE, cpu_readmem24_dword(0x68L)); // $68 *must* be copied at 10F6EE
	cpu_writemem24(0x00013B, 0x01);
}	

void	neogeo_reset(void)
{
	memset((void *)&regs, 0, sizeof(regs));
	regs.pc = 0x122;
	regs.sr = 0x2700;
	regs.sr_h = 0x27;
	regs.a7 = regs.isp = regs.usp = 0x10F300;
	OP_ROM = neogeo_prg_memory;
	neogeo_vline = 0;
	neogeo_68k_cycles = CYCLES_PER_LINE;
	cpu_writemem24(0x10FD80, 0x82);
	cpu_writemem24(0x10FDAF, 0x01);
	cpu_writemem24(0x10FEE1, 0x0A);
	cpu_writemem24(0x10F675, 0x01);
	cpu_writemem24(0x10FEBF, 0x00);
	cpu_writemem24_dword(0x10FDB6, 0);
	cpu_writemem24_dword(0x10FDBA, 0);
}

void	neogeo_shutdown(void)
{
	FILE	*fp;
	
	video_set_mode(VIDEO_TEXT);
	puts("NEOGEO: System Shutdown.");

	if ((windows_present)&&(windows_hostile_mode))
		winlow_enable_windows();

	unload_datafile(rsc);
	remove_keyboard();
	remove_mouse();
	remove_timer();
	cdda_shutdown();
	cdrom_shutdown();
	video_shutdown();

	fp = fopen("memcard.bin", "wb");
	fwrite(neogeo_memorycard, 1, 8192, fp);
	fclose(fp);

	free(neogeo_prg_memory);
	free(neogeo_rom_memory);
	free(neogeo_spr_memory);
	free(neogeo_fix_memory);
}

void	neogeo_exception(void)
{
	video_set_mode(VIDEO_TEXT);
	printf("NEOGEO: Exception Trapped at %08x !\n", previouspc);

	exit(0);
}	

void MC68000_Cause_Interrupt(int level)
{
	if (level >= 1 && level <= 7)
        regs.irq |= 1 << (level-1);
}

void	neogeo_exit(void)
{
	video_set_mode(VIDEO_TEXT);
	puts("NEOGEO: Exit requested by software...");
	exit(0);
}

void	neogeo_run(void)
{
	if (!neogeo_ipl_done)
	{
		cdrom_load_title();

		// Process IPL.TXT
		if (!cdrom_process_ipl()) {
			neo_alert1("Error", NULL, "Could not open IPL.TXT.", NULL, "&Abort", 'a');
			return;
		}
		
		neogeo_ipl_done = 1;
		neogeo_hreset();
	}

	remove_mouse();
	remove_timer();
	
	text_mode(0);
	
	if (config_scanlines)
		video_set_mode(VIDEO_SCANLINES);
	else
		clear(screen);
		
	while(!key[KEY_TILDE])
	{
		MC68000_ICount = 0;
		M68KRUN(CYCLES_PER_LINE);

		neogeo_68k_cycles += MC68000_ICount;
		if (neogeo_68k_cycles < 0) {
			neogeo_vline++;
			neogeo_68k_cycles += CYCLES_PER_LINE;
			if (neogeo_vline==224) {
				neogeo_vline=0;

				if (key[KEY_F11])
				{
					video_hide_fps ^= 1;
					while(key[KEY_F11]);
				}

				poll_joystick();
	
				if (neogeo_prio_mode)
					video_draw_screen2();
				else
					video_draw_screen1();

				if (key[KEY_F12])
					video_save_snapshot();

				MC68000_Cause_Interrupt(2);
			}
		}
	}

	clear_keybuf();
	install_timer();
	install_mouse();
	video_set_mode(VIDEO_NORMAL);

	set_mouse_sprite(rsc[CURSOR].dat);
	show_mouse(screen);
}


// This is a really dirty hack to make SAMURAI SPIRITS RPG work

void	neogeo_prio_switch(void)
{
//	video_set_txt_mode();
//	debug_dump_regs();
//	clear_keybuf();
//	readkey();
//	video_set_gfx_mode();
	
	if (regs.d7 == 0xFFFF)
		return;
	
	if ((regs.d7==9)&&(regs.a3==0x10DED9)&&((regs.a2==0x1081d0)||((regs.a2&0xFFF000)==0x102000)))
	{
		neogeo_prio_mode = 0;
		return;
	}
	
	if ((regs.d7==8)&&(regs.a3==0x10DEC7)&&(regs.a2==0x102900))
	{
		neogeo_prio_mode = 0;
		return;
	}
	
	if (regs.a7 == 0x10F29C)
	{
		if ((regs.d4&0x4010)==0x4010)
		{
			neogeo_prio_mode = 0;
			return;
		}
		
		neogeo_prio_mode = 1;
	}
	else
	{
		if (regs.a3==0x5140)
		{
			neogeo_prio_mode = 1;
			return;
		}

		if ( (regs.a3&~0xF) == (regs.a4&~0xF) )
			neogeo_prio_mode = 1;
		else
			neogeo_prio_mode = 0;
	}
}

void	draw_main(void)
{
	int	i;
	
	scare_mouse();
	
	blit(rsc[MAIN].dat, screen, 0, 0, 0, 0, 320, 240);
	
	if (neogeo_ipl_done)
		blit(game_title, screen, 0, 0, 35, 30, 144, 80);

	for(i=0;i<NB_ELEMS;i++)
		blit(rsc[main_buttons[i].nrm].dat, screen, 0, 0, main_buttons[i].x1, main_buttons[i].y1, 71, 13);

	unscare_mouse();
}	

void	not_implemented(void)
{
		neo_alert1("Error", NULL, "This function isn't implemented.", NULL, "&Abort", 'a');
}

void	neogeo_quit(void)
{
	
	if (neo_alert2("Confirmation", NULL, "Are you sure ?", NULL, "&Yes", "&No", 'y', 'n') == 0)
	{
		video_set_mode(VIDEO_TEXT);
		exit(0);
	}
}

void neogeo_machine_settings(void)
{
	int		ret, temp;
	FILE	*fp;

	centre_dialog(machine_settings);

	ret = get_config_int(NULL, "cdrom", -1);

	machine_settings[9].d1 = 0;
	
	if (ret != -1)
	{
		for(temp=0;temp<nb_of_drives;temp++)
		{
			if (drive_list[temp] == ret)
				machine_settings[9].d1 = temp + 1;
		}
	}

	temp = get_config_int(NULL, "nationality", 0);
	if (temp>2)
		temp = 2;
	machine_settings[4].d1 = temp;

	do
	{
		vsync();
		ret = do_dialog(machine_settings, 4);

		if (machine_settings[4].d1 != temp)
		{
			fp = fopen("startup.bin", "rb");
			if (fp==NULL)
				neo_alert1("Error", NULL, "Could not load STARTUP.BIN.", NULL, "&Abort", 'a');
			else
			{
				temp = machine_settings[4].d1;
				set_config_int(NULL, "nationality", temp);
				fread(neogeo_prg_memory + 0x10F300, 1, 3328, fp);
				fclose(fp);
				swab(neogeo_prg_memory + 0x10F300, neogeo_prg_memory + 0x10F300, 3328);
				cpu_writemem24(0x10FD83, temp);
				neogeo_ipl_done = 0;
			}
		}

		if (ret == 5)
			neogeo_reset();
		
		if (ret == 6)
			neogeo_ipl_done = 0;
			
		if (ret == 11)
		{
			if (machine_settings[9].d1 == 0)
			{
				cdrom_current_drive = drive_list[0];
				set_config_int(NULL, "cdrom", -1);
			}
			else
			{
				cdrom_current_drive = drive_list[machine_settings[9].d1 - 1];
				set_config_int(NULL, "cdrom", drive_list[machine_settings[9].d1 - 1]);
			}
			
			neogeo_ipl_done = 0;
		}
	
		if (ret == 13)
		{
			about_box[1].dp = rsc[NGCD].dat;
			about_box[1].dp2 = rsc[NGCD].dat;
			centre_dialog(about_box);
			popup_dialog(about_box, 0);
		}			
	} while( ret != 14 );
}
